import datetime
import marshmallow_dataclass
from typing import List, Dict, Any, Optional, Union
from dataclasses import field
from marshmallow import Schema, post_dump, validate, INCLUDE
from .common import BaseEnum

# --- SERIALIZERS -------------------------------------------------------------


class BaseSchema(Schema):
    SKIP_VALUES = [None, [], {}]

    class Meta:
        ordered = True
        unknown = INCLUDE

    @post_dump
    def remove_skip_values(self, data, **kwargs):
        return self._remove_skip_values(data)

    @classmethod
    def _remove_skip_values(cls, obj):
        if isinstance(obj, list):
            return [*map(cls._remove_skip_values, obj)]
        elif isinstance(obj, dict):
            return {k: cls._remove_skip_values(v) for k, v in obj.items() if v not in cls.SKIP_VALUES}
        return obj


class Measures(BaseEnum):
    """Used with waterfall"""
    RELATIVE = 'relative'
    ABSOLUTE = 'absolute'
    TOTAL = 'total'


class Textpositions(BaseEnum):
    """Used with waterfall"""
    INSIDE = "inside"
    OUTSIDE = "outside"
    AUTO = "auto"
    NONE = "none"


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class LinkModel:
    source: List[Any] = field(default_factory=list)  # indices correspond to NodeModel.label
    target: List[Any] = field(default_factory=list)
    value: List[Any] = field(default_factory=list)
    label: List[str] = field(default_factory=list)
    color: List[str] = field(default_factory=list)
    customdata: Optional[List[str]] = field(default_factory=list)
    hovertemplate: str = field(default=None)


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class NodeModel:
    x: List[float] = field(default_factory=list)
    y: List[float] = field(default_factory=list)
    label: List[str] = field(default_factory=list)
    color: Optional[Union[List[str], str]] = field(default=None)
    customdata: Optional[List[str]] = field(default_factory=list)
    pad: Optional[int] = field(default=None)
    thickness: Optional[int] = field(default=None)
    line: Optional[Dict[str, Any]] = field(default=None)
    hovertemplate: str = field(default=None)


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class MarkerModel:
    opacity: Optional[float] = field(default=None)
    symbol: Optional[List[int]] = field(default=None)
    color: Optional[str] = field(default=None)
    colors: Optional[List[str]] = field(default_factory=list)


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class TableFormatModel:
    style: Optional[Dict[str, Any]] = field(default_factory=dict)
    type: Optional[str] = field(default=None)
    format: Optional[str] = field(default=None)


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class ChartDataModel:
    x: List[Any] = field(default_factory=list)
    y: List[Any] = field(default_factory=list)
    labels: List[Any] = field(default_factory=list)
    values_: List[Any] = field(default_factory=list, metadata=dict(data_key='values'))
    parents: List[Any] = field(default_factory=list)
    hovertext: List[Any] = field(default_factory=list)
    ids: List[Any] = field(default_factory=list)
    measure: List[str] = field(default_factory=list, metadata=dict(validate=validate.ContainsOnly(Measures.to_list())))
    text: List[str] = field(default_factory=list)
    name: str = field(default=None)
    branchvalues: Optional[str] = field(default=None)
    link: Optional[LinkModel] = field(default=None)
    node: Optional[NodeModel] = field(default=None)
    hoverinfo: Optional[str] = field(default=None)
    hovertemplate: Optional[str] = field(default=None)
    textinfo: Optional[str] = field(default=None)
    hole: Optional[float] = field(default=None)
    line: Optional[Dict[str, Any]] = field(default=None)
    totals: Optional[Dict[str, Any]] = field(default=None)
    increasing: Optional[Dict[str, Any]] = field(default=None)
    decreasing: Optional[Dict[str, Any]] = field(default=None)
    marker: Optional[MarkerModel] = field(default=None)
    fill: Optional[str] = field(default=None)
    mode: Optional[str] = field(default=None)
    orientation: Optional[str] = field(default=None)
    arrangement: Optional[str] = field(default=None)
    valueformat: Optional[str] = field(default=None)
    textposition: Optional[str] = field(default=None, metadata=dict(validate=validate.OneOf(Textpositions.to_list())))
    base: Optional[int] = field(default=None)
    showlegend: Optional[bool] = field(default=None)
    type: Optional[str] = field(default=None)  # only for use of backwards plotly regeneration
    headers: Optional[List[str]] = field(default_factory=list)
    is_header_row: Optional[bool] = field(default=None, metadata=dict(data_key='isHeaderRow'))
    format: Optional[TableFormatModel] = field(default=None)
    context: Optional[str] = field(default=None)
    stratification: Dict[str, Any] = field(default_factory=dict)
    domain: Optional[Dict[str, Any]] = field(default=None)


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class ChartModel:
    type: str = field(init=True, metadata=dict(required=True))  # Define the plotly chart type, such as bar, pie, etc.
    id: str = field(default='chart1')  # Optional?
    data: List[ChartDataModel] = field(default_factory=list, repr=False, metadata=dict(validate=validate.Length(min=1)))
    mode: Optional[str] = field(default=None)
    data_count: Optional[int] = field(default=None, metadata=dict(data_key='dataCount'))
    format: Optional[TableFormatModel] = field(default=None)


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class LayoutModel:
    title: Union[str, Dict[str, Any]] = field(default='', metadata=dict(required=False))
    xaxis: Dict[str, Any] = field(default_factory=lambda: dict(visible=True, showline=True))
    yaxis: Dict[str, Any] = field(default_factory=lambda: dict(visible=True, showline=True))
    font: Optional[Dict[str, Any]] = field(default_factory=dict)
    titlefont: Optional[Dict[str, Any]] = field(default_factory=lambda:
        dict(color='rgb(22, 96, 167)', size=20))
    plot_bgcolor: str = field(default='rgb(255,255,255)')
    annotations: List[Any] = field(default_factory=list)
    height: int = field(default=700)
    width: int = field(default=1200)
    autosize: bool = field(default=True)
    barmode: Optional[str] = field(default=None)
    margin: Optional[Dict[str, Any]] = field(default_factory=dict)
    shapes: Optional[List[Dict[str, Any]]] = field(default_factory=list)
    updatemenus: Optional[List[Dict[str, Any]]] = field(default_factory=list)
    waterfallgroupgap: Optional[float] = field(default=None)

    def convert_for_plotly(self, **kwargs: Any):
        for key, value in kwargs.items():
            if key in self.__dict__:
                self.__dict__[key] = value
        self.autosize = False
        return self

    def update_axes(self, **kwargs: dict):
        self.xaxis.update(kwargs)
        self.yaxis.update(kwargs)

    def add_axes_titles(self, xaxis_title: str, yaxis_title: str):
        self.xaxis['title'], self.yaxis['title'] = xaxis_title, yaxis_title

    def set_data_points(self, data_len: int):
        if data_len > 500:
            self.xaxis['dtick'] = 100
        elif data_len > 50:
            self.xaxis['dtick'] = 10


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class ReportModel:
    id: str = field(default='report1')
    name: str = field(default='')
    layout: LayoutModel = field(default=None, metadata=dict(validate=validate.Length(min=1)))
    charts: List[ChartModel] = field(default_factory=list, metadata=dict(validate=validate.Length(min=1)))
    strata_label: str = field(default=None, metadata=dict(data_key='stratificationLabel'))


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class StratumItemModel:
    id: int = field(default=0)
    name: str = field(default='')
    parent_id: int = field(default=None, metadata=dict(data_key='parentId'))


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class StratificationModel:
    id: str = field(default='')
    name: str = field(default='')
    items: List[StratumItemModel] = field(default_factory=list, repr=False)


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class VisualisationModel:
    reports: List[ReportModel] = field(default_factory=list, metadata=dict(validate=validate.Length(min=1)))
    stratifications: List[StratificationModel] = field(default_factory=list)
    id: str = field(default=None, metadata=dict(load_only=True))
    status: Any = field(default=None, metadata=dict(load_only=True))
    message: str = field(default=None, metadata=dict(load_only=True))
    version: int = field(default=None, metadata=dict(load_only=True))
    created: datetime.datetime = field(default=None, metadata=dict(load_only=True))
    updated: datetime.datetime = field(default=None, metadata=dict(load_only=True))


@marshmallow_dataclass.dataclass(base_schema=BaseSchema)
class _PlotlyPayloadModel:
    """
    Private model for mapping to plain Plotly objects
    """
    data: List[ChartDataModel]
    layout: LayoutModel
